import fs from 'fs';
import webpack from 'webpack';
import { merge } from 'webpack-merge';
import { Options } from './start';

import getOutput from './getOutput';
import getLoaders from './getLoaders';
import getPlugins from './getPlugins';

import getOptimization from './getOptimization';
import mockMiddleware from './middleware/mock';
import loggerMiddleware from './middleware/logger';
import { RcbConfig } from './getConfig';
import utils from './utils';
import paths from './paths';

const entry = ['./src/index'];

export interface Configuration extends webpack.Configuration {
    devServer?: any
}

function getExtraWebpackConfig(webpackConfig: Configuration) {
    let path: string = '';
    if (utils.isDev()) {
        path = paths.getResolvePath('webpack.config.dev.js');
    }

    if (utils.isProduction()) {
        path = paths.getResolvePath('webpack.config.prod.js');
    }

    if (!(path && fs.existsSync(path))) {
        const webpackConfig = paths.getResolvePath('webpack.config.js');

        if (webpackConfig && fs.existsSync(webpackConfig)) {
            path = webpackConfig;
        }
    }

    if (!path) {
        return {};
    }

    try {
        const config = utils.requireModule(path);
        if (typeof config === 'function') {
            return config(webpackConfig);
        }
        return config;
    }
    catch (e) {
        return {};
    }
}

const HOST = '0.0.0.0';
const PORT = 3000;

export default (rcb: RcbConfig, options?: Options): webpack.Configuration => {
    const devServerConfig: any = {
        host: rcb.host || HOST,
        port: rcb.port || PORT,
        proxy: rcb.proxy,
        historyApiFallback: {
            htmlAcceptHeaders: ['text/html', 'application/xhtml+xml', 'application/xml']
        },
        compress: true,
        client: {
            logging: "error",
            overlay: false
        },
        allowedHosts: 'all',
        open: false,
        devMiddleware: {
            stats: false
        },
        setupMiddlewares(middlewares: any, devServer: any) {
            const mockConfig = rcb.mock;
            if (options?.mock && mockConfig) {
                // mock代理的url不会在node中log出来，只有proxy代理的url才会log方便查看是哪里出了问题
                devServer.app.use(mockConfig.api, mockMiddleware.middleware(rcb.mock));
            }
            middlewares.push(loggerMiddleware());

            return middlewares;
        }
    };

    const defaultWebpackConfig: webpack.Configuration & { devServer?: any } = {
        mode: utils.getEnv(),
        performance: {
            hints: false
        },
        stats: false,
        devServer: utils.isDev() ? Object.assign(devServerConfig, rcb.devServer) : undefined,
        devtool: utils.isDev() && rcb.devtool ? rcb.devtool : false,
        entry: rcb.entry || entry,
        infrastructureLogging: {
            level: 'error'
        },
        output: getOutput(rcb.output),
        module: {
            rules: getLoaders(rcb)
        },
        optimization: getOptimization(rcb),
        plugins: getPlugins(rcb),
        resolve: {
            modules: ['node_modules', paths.appNodeModules].concat(
                rcb.additionalModulePaths || []
            ),
            alias: rcb.alias,
            extensions: ['.js', '.jsx', '.es6', '.json', '.ts', '.tsx', '.mjs'],
            fallback: {
                "fs": false,
                "tls": false,
                "net": false,
                "path": false,
                "zlib": false,
                "http": false,
                "https": false,
                "stream": false,
                "crypto": false
            }
        },
        watchOptions: utils.isDev() ? {
            poll: 100,
            ignored: ["node_modules"]
        } : undefined
    };

    const config = merge(
        defaultWebpackConfig,
        getExtraWebpackConfig(defaultWebpackConfig),
    );

    return merge(
        config,
        typeof rcb.webpack === 'function'
            ? rcb.webpack(config)
            : rcb.webpack || {}
    );
};
